#define _FILE_OFFSET_BITS 64
#define _LARGEFILE_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ctype.h>

#ifdef __MSVCRT__
#define ftello ftello64
#define fseeko fseeko64
#define off_t off64_t
#define mkdir(name,mode) _mkdir(name)
#endif

#define MAX_CALLS 200

typedef int bool;
#define true 1
#define false 0

// originally assumed to be 20, insufficient for Blue Dragon
#define SECTORS_PER_LEVEL 40

// 20 sublevels max * X sectors each
static char buffer[20*2048*SECTORS_PER_LEVEL];
static char dirprefix[PATH_MAX+1] = {0};
static char currdir[PATH_MAX+1] = {0};
static char newdir[PATH_MAX+1] = {0};
static bool doWildcard = false;
static char strWildcard[100] = {0};
static off_t totalbytes = 0;
static off_t baseoffset = 0;
static char ripdirname[PATH_MAX+1] = {0};
static char rootdirname[PATH_MAX+1] = {0};

int readBlock(FILE *fp, off_t LBA, unsigned char* data, unsigned short num);
int makeint(char in);
void printentry(char *entry, off_t offset, char* dirprefix);
void printprogress(int current, int last, int total);
void parse( char *entry , off_t offset, int level, FILE* fp, bool rip, char* dirprefix);
void extract( char *entry, off_t offset, FILE* fp);
void extractFile( char *entry, off_t offset, FILE* fp, char* dirprefix);
int match(const char *mask, const char *name);
bool fnameMatches( char *entry, off_t offset);
void multicreate(char *dirname);
unsigned long getint(char*);
unsigned short getword(char*);
void init(void);

int main(int argc, char* argv[])
{
    FILE *fp;
    unsigned long rootsector;
    unsigned long rootsize;
	bool doexit = false;

    printf("***********************************\n");
    printf("** PI XBOX-360 DUMP EXTRACT V1.2 **\n");
	printf("**    Xtreme Mod by Redline99    **\n");
	printf("**    otherwise modded by hcs    **\n");
    printf("***********************************\n\n");

    if ( argc < 2 )
    {
        printf("Usage: %s gamename.360 <wildcard>...\n\n",argv[0]);
        return 1;
    }

    fp = fopen(argv[1], "rb");
    if (fp==0) 
    {
        printf(" * Error, could not open the file %s\n", argv[1]);
        return 1;
    }

    if ( argc == 3 )
    {
        // wildcard
        doWildcard = true;

        // copy wildcard string
        strcpy(strWildcard, argv[2]);

        printf(" Doing wildcard extraction, wildcard is: \"%s\"\n\n", strWildcard);
    }

    // now we are ready to read.. first let's see if this is a game partition only by reading sector 32
    if (readBlock(fp, 32, (unsigned char*) buffer, 1)) exit(1);
    if (strncmp(buffer, "MICROSOFT*XBOX*MEDIA", 0x14) != 0)
    {
		baseoffset = 129824; //Xbox 360 Xtreme
		if (readBlock(fp, baseoffset + 32, (unsigned char*) buffer, 1)) exit(1);
		if (strncmp(buffer, "MICROSOFT*XBOX*MEDIA", 0x14) != 0)
		{
			baseoffset = 198144; //Xbox One Xtreme
			if (readBlock(fp, baseoffset + 32, (unsigned char*) buffer, 1)) exit(1);

			if (strncmp(buffer, "MICROSOFT*XBOX*MEDIA", 0x14) != 0)
			{
				printf(" * Error, This is not a XBOX DVD image! Exiting...\n");
				doexit = true;
			}
			else
			{
				printf("Xtreme style Xbox One image detected\n");
			}
		}
		else
		{
			printf("Xtreme style Xbox 360 image detected\n");
		}
	}
	else
	{
		printf("Game partition only Xbox image detected\n");
    }
	
	if (doexit == true) 
	{
		return 1;
	}


    printf("Detecting... relative ROOT-Sector = ");
    rootsector = getint(buffer+0x14);
    rootsize = getint(buffer+0x18);
    printf("%lu (0x%lx), length is %lu (0x%lx)\n", rootsector,rootsector, rootsize, rootsize);

    // now we have the rootsector address.. we should read it now
    if (readBlock(fp, baseoffset + rootsector, (unsigned char*)buffer, SECTORS_PER_LEVEL)) exit(1);

    // print TOC
    printf("\n");
    printf("sector        length   filename\n");
    printf("---------------------------------------------------------------------------\n");

    init();
    parse(buffer, 0, 0, fp, false, dirprefix);
    printf("\nTotal bytes used on this XBOX dvd: %lld\n", totalbytes);

    printf("\n\nNow extracting files\n");
    printf("====================\n\n");

    // extract all files
    int dirpostfix = 0;
    bool done = false;

    while (!done)
    {
        sprintf(ripdirname, "PI_XBOX360_%u", dirpostfix);
        if (mkdir(ripdirname,0744) == 0)
            done = true;
        else
        {
            dirpostfix++;
            memset(ripdirname, 0, 20);
        }
    }
    
    getcwd(rootdirname, PATH_MAX);

    // now that we found a new dirname, change to it
    chdir(rootdirname);
    chdir(ripdirname);


    // finally we can rip all files...
    init();
    parse(buffer, 0, 0, fp, true, dirprefix);

    printf("\nAll done! Enjoy ;-)\n");

    return 0;
}

int match(const char *mask, const char *name)
{
	int calls=0, wild=0, q=0;
	const char *m=mask, *n=name, *ma=mask, *na=name;

	for(;;) 
	{
		if (++calls > MAX_CALLS) 
			return 1;
			if (*m == '*') 
			{
				while (*m == '*') ++m;
					wild = 1;
					ma = m;
					na = n;
			}

		if (!*m) 
		{
			if (!*n) return 0;

			for (--m; (m > mask) && (*m == '?'); --m) ;

			if ((*m == '*') && (m > mask) && (m[-1] != '\\')) 
				return 0;
			if (!wild) 
				return 1;
			m = ma;
		} 
		else if (!*n) 
		{
			while(*m == '*') ++m;
			return (*m != 0);
		}
		if ((*m == '\\') && ((m[1] == '*') || (m[1] == '?'))) 
		{
			++m;
			q = 1;
		} 
		else 
		{
			q = 0;
		}

		if ((tolower(*m) != tolower(*n)) && ((*m != '?') || q)) 
		{
			if (!wild) 
				return 1;
			m = ma;
			n = ++na;
		} 
		else 
		{
			if (*m) ++m;
			if (*n) ++n;
		}
	}
}

void multicreate(char dirname[])
{
    if (dirname[0] == 0)
        return;

    char pir[100] = {0};

    int currpos = 0;
    char* pos = strchr(dirname, '/');

    memset(pir, 0, 100);
    strncpy(pir, dirname+currpos, pos-dirname);
    mkdir(pir,0744);
    chdir(pir);

    if (strchr(pos+1, '/') == 0)
        return;

    currpos = pos-dirname+1;
    while(pos != 0)
	{
		dirname = pos+1;
        pos = strchr(dirname, '/');
        memset(pir, 0, 100);
        if (pos == 0)
            strcpy(pir, dirname);
        else
            strncpy(pir, dirname, pos-dirname);

        mkdir(pir,0744);
        chdir(pir);
    }
        
}

void parse( char *entry, off_t offset, int level, FILE* fp, bool rip, char* dirprefix)
{
	if (offset>=SECTORS_PER_LEVEL*2048) {
		printf("]>> SECTORS_PER_LEVEL is too low!! <<[\n");
		return;
	}
    //printf("parse (offset=%llx, level=%d)\n",offset,level);
    // THERE SEEM TO BE GAMES THAT HAVE INVALID TOC ENTRIES IN THEM!!
	unsigned long sector = getint(entry+offset+4);
    if (sector == 0xFFFFFFFF)
    {
        if (rip)
        {
            chdir(rootdirname);
            chdir(ripdirname);
            multicreate(dirprefix);
        }
        // invalid shit!!
        printf("]>> INVALID TOCENTRY DETECTED! <<[\n");
        return;
    }

    unsigned short left = getword(entry+offset);
    unsigned short right = getword(entry+offset+2);

    if (left)
        parse(entry, left*4, level, fp, rip, dirprefix);

    if (*(entry+offset+0x0c) == 0x10 || *(entry+offset+0x0c) == 0x30 /* 0x30? yeah, 0x30! */)
    {
        // directory
        unsigned long tocsector = getint(entry+offset+4);

        char dirname[PATH_MAX+1];
        int dirnamesize = 0;

        dirnamesize = makeint(*(entry+offset+0xd));
        strncpy(dirname, entry+offset+0x0e, dirnamesize);
        dirname[dirnamesize] = 0x0;

        char newdirprefix[PATH_MAX+1] = {0};
        if (*dirprefix == 0)
            strcpy(newdirprefix, dirname);
        else
            sprintf(newdirprefix, "%s%s", dirprefix, dirname);
        strcat(newdirprefix, "/");

        // special!! a sector is not mandatory! (empty directories...)
        if ( (tocsector == 0) && rip)
        {
            // no files below this, so just create the directory
            strcpy(newdir, dirprefix);
            chdir(rootdirname);
            chdir(ripdirname);
            multicreate(dirprefix);
        }
        else
        {
            if (readBlock(fp, baseoffset +  tocsector, (unsigned char*)(buffer+(level+1)*2048*SECTORS_PER_LEVEL), SECTORS_PER_LEVEL)) {
		exit(1);
	    }
            parse(buffer+((level+1)*2048*SECTORS_PER_LEVEL), 0, level+1, fp, rip, newdirprefix);
        }
    }
    else
    {
        strcpy(newdir, dirprefix);

        if (rip)
        {
            if ( (doWildcard && fnameMatches(entry, offset)) || !doWildcard )
            {
                // we here should create the new dir and descend into it
                chdir(rootdirname);
                chdir(ripdirname);
                multicreate(dirprefix);
    
                extractFile(entry, offset, fp, dirprefix);
            }
        }
        else
        {
            printentry(entry, offset, dirprefix);
        }
    }
    if (right)
        parse(entry, right*4, level, fp, rip, dirprefix);
}

void printentry(char *entry, off_t offset, char* dirprefix)
{
    // print sector
    unsigned long sector = getint(entry+offset+4);
    printf("%07lu   ", sector);
    unsigned long size = getint(entry+offset+8);
    totalbytes+=size;
    printf("%10lu   ", size);

    if (*dirprefix != 0)
        printf("%s", dirprefix);
    printf("%.*s\n", makeint(*(entry+offset+0xd)), entry+offset+0x0e);
}

void printprogress(int current, int last, int total)
{
    int numleft = last - current;
    int numdone = total - numleft;
    int counter = numdone * 100;
    int process = counter/total;
    printf("\b\b\b\b");
    printf("%3u%%", process);

}

bool fnameMatches( char *entry, off_t offset)
{
    char filename[PATH_MAX+1] = {0};
    int fnamesize = 0;
    fnamesize = makeint(*(entry+offset+0xd));
    strncpy(filename, entry+offset+0x0e, fnamesize);
    filename[fnamesize] = 0x0;

    if (match(strWildcard, filename) == 0)
        return true;
    else
        return false;
    return false;
}

void extractFile( char *entry, off_t offset, FILE* fp, char* dirprefix)
{
    char filename[PATH_MAX+1] = {0};
    int fnamesize = 0;
    char filebuffer[2048];
    //char bogobuffer[2048] = {0};

    fnamesize = makeint(*(entry+offset+0xd));
    strncpy(filename, entry+offset+0x0e, fnamesize);

    filename[fnamesize] = 0x0;

    if (strcmp(currdir, newdir) != 0)
    {
        if ( *newdir == 0 )
            printf("\n ==> Processing directory: /\n");
        else
            printf("\n ==> Processing directory: %s\n", newdir);
        strcpy(currdir, newdir);
    }

    printf("Ripping '%s' ", filename);
    unsigned long sector = getint(entry+offset+4);
    unsigned long size = getint(entry+offset+8);
    unsigned long numsectors = size / 0x800;                // number of sectors to read
    unsigned long restbytes = size - (numsectors*0x800);

    FILE *ripfile;
    ripfile = fopen(filename, "wb");
    //bool error = false;
    bool error = (ripfile==NULL);
    //int iTry = 0;

    printf("[sector: %lu (%lx) - size: %lu (%lx) bytes]...     ", sector, sector, size, size);

    // now let's see
    {
    int i;
    for (i = sector; (unsigned)i < sector+numsectors && !error; i++) // skip last sector, we will read manually
    {
        if (readBlock(fp, baseoffset + i, (unsigned char*)filebuffer, 1) != 0)
            error = true;
        else
        {
            // print progress bar stuff
            printprogress(i, sector+numsectors-1, numsectors);
        
            fwrite(filebuffer, 2048, 1, ripfile);
        }

        if (error)
        {
            printf(" ...Error at sector %u! Skipping!\n", i);
            break;
        }
    }
    }

    // get the restbytes
    if (!error && restbytes != 0)
    {
        if (readBlock(fp, baseoffset + sector + numsectors, (unsigned char*)filebuffer, 1) != 0)
            error = true;
        else
        {
            // print progress bar stuff
            printprogress(1, 1, 1);
            fwrite(filebuffer, restbytes, 1, ripfile);
        }
    }

    if (error)
    {
        printf(" ...Error at sector %lu! Skipping!\n", sector + numsectors);
        fclose(ripfile);
        return;
    }

    printf(" ...Done!\n");

    // now let's see
    fclose(ripfile);
}

int readBlock(FILE *fp, off_t LBA, unsigned char* data, unsigned short num)
{
    //fprintf(stderr,"read %d at %llx (sector %llu)\n",num,LBA*2048,LBA);
    if(fseeko(fp, LBA*2048, SEEK_SET)) {
	printf(" * error seeking to 0x%llx (block %llu)\n",LBA*2048,LBA);
	return 1;
    }
    if( fread(data, 1, num*2048, fp) != (unsigned)num*2048 ) {
	printf(" * error reading %d bytes (%d blocks) from 0x%llx (block %llu)\n",num*2048,num,LBA*2048,LBA);
	return 1;
    }
    return 0;

}

int makeint(char in)
{
    int result = 0;
    result = in;
    return result;
}

unsigned short getword(char* ptr)
{
    unsigned short ret;
    ret = (((unsigned char) *(ptr+1)) & 0xFF) << 8;
    ret |= (((unsigned char) *ptr)) & 0xFF;
    
    return ret;
}

unsigned long getint(char* ptr)
{

    unsigned long ret;
    ret = (*(ptr+3) & 0xFF) << 24;
    ret |= (*(ptr+2) & 0xFF) << 16;
    ret |= (*(ptr+1) & 0xFF) << 8;
    ret |= *(ptr+0) & 0xFF;

    return ret;
}

void init(void) {
	memset(dirprefix,0,sizeof(dirprefix));
	memset(currdir,0,sizeof(currdir));
	memset(newdir,0,sizeof(newdir));
	currdir[0] = '/';
	totalbytes=0;
}
